# Dynamic Fast Marching for arbitrary dimension. SIAM version
using DataStructures, LinearAlgebra #
using Plots
Plots.gr(fontfamily="Times")
################################################################ 
include("auxfunctions.jl")

struct Grid
    T :: Array{Float64} # time values
    S :: Array{Bool} # status: true for fixed value
end


# Dynamic fast marching method 3D# 
# change to time_p to use data from the past
function dfmm(;obstacle=(_,_)->1,initSet,n=nmax)
    dim = length(initSet[1])
    initSet = [pt2idx(p) for p in initSet]
    dims = Tuple((n+1) * ones(Int,dim))
    H = MutableBinaryMinHeap{Tuple{Float64,CartesianIndex{dim}}}() # data structure to implement 
    # the "Heap". First value is time, second, the pixel coordinates 
    G = Grid(fill(Inf,dims),falses(dims)) # initial time values in a Matrix of n nodes in each dimension
    V = falses(dims)
    handle = Dict{CartesianIndex{dim},Int}() # stores the handle of the pixel pointing to H
    for p in initSet
        handle[p] = push!(H,(0.,p)) # puts the initial values in H
    end
    # main loop
    while !isempty(H)
        # updating the Heap and handles using the pixel with least arriving time
        t,p = pop!(H) # takes the minimum time value and the pixel coordinates in narrow band
        G.S[p] = true # accepted value
        G.T[p] = t # assign time even if it is in the obstacle
        delete!(handle,p) # deleted from the dictionary of handles, already accepted
        if  obstacle(idx2pt(p),t/n) > 0 # out of the obstacle neighbors are added
            # assign tentative times to the neighbors that are not fixed
            for q in [v for v in neighbors(p) if !G.S[v]]
                t_ = compute_time(p,q,G) # time_classic 
                if haskey(handle,q) # in this case it was visited before (it could have been visited and have an infinite value)
                    h = handle[q]
                    if t_ < H[h][1] # the new tentative value is lower than the one stored
                        DataStructures.update!(H, h, (t_,q)) # updates value
                    end
                else # not visited before (not accepted and not being processed)  
                    handle[q] = push!(H,(t_,q)) # added to the narrow band
                end                        
            end
        else
            V[p] = true
        end
    end
    return G, V #G.T/n 
end


################### time functions #######################
# Dimension dependent methods
const sq2,sq3 = sqrt(2),sqrt(3)

function compute_time(p::CartesianIndex{3},q::CartesianIndex{3},G::Grid)
    t = Inf
    for i in 0:2
        n  = base3[(i)   % 3 + 1]
        e1 = base3[(i+1) % 3 + 1]
        e2 = base3[(i+2) % 3 + 1]
        for s = [-1,1] # for each direction, two normals. We can use the same e1,e2 for both faces
            c = q + s * n
            t1 = filter_time(p,c,G)
            for d1 in [e1,-e1], d2 in [e2,-e2]
                t2  = min(filter_time(p,c+d1,G),filter_time(p,c+d2,G)) # select the optimal in a single square
                t3  = filter_time(p,c+d1+d2,G) # the same for the two triangles in the face
                t   = min(t,local_solver((t1,t2,t3)))
            end
        end
    end
    return t
end

#########################################################################
function compute_time(p::CartesianIndex{2},q::CartesianIndex{2},G::Grid)
    t = Inf
    for i in 0:1
        n  = base2[(i)   % 2 + 1]
        e1 = base2[(i+1) % 2 + 1]
        for s = [-1,1] # for each direction, two normals. We can use the same e1,e2 for the face
            c  = q + s * n
            t1 = filter_time(p,c,G)
            t2 = min(filter_time(p,c+e1,G),filter_time(p,c-e1,G)) # select the optimal in a single segment
            t  = min(t,local_solver((t1,t2)))
        end
    end
    return t
end

# checks that there is a finite accepted value and that satisfies the constraints for node q (neighbor of p)
function filter_time(p::CartesianIndex{3},q::CartesianIndex{3},G::Grid)
    in_domain(q) && G.S[q] && G.T[q] >= G.T[p] - sq3 ? G.T[q] : Inf # 
end

function filter_time(p::CartesianIndex{2},q::CartesianIndex{2},G::Grid)
    in_domain(q) && G.S[q] && G.T[q] >= G.T[p] - sq2 ? G.T[q] : Inf # ξ := sqrt(3) # 
end

#######################################################################
function local_solver(t::NTuple{2})
    Δ = t[1]-t[2]
    a = f1(t[1])
    b = f2(t[2])
    c = f12(t[1],Δ)
    return min(a,b,c)
end

function local_solver(t::NTuple{3})
    Δ = t[1] - t[2]
    H = t[2] - t[3]
    M = t[1] - t[3]
    a = t[1] + 1.
    b = t[2] + sq2
    c = t[3] + sq3
    d = f12(t[1],Δ)
    e = f23(t[2],H)
    f = f13(t[1],M)
    g = f123(t[1],Δ,H)
    return min(a,b,c,d,e,f,g)
end

####################################################
# zero dimensional manifolds
function f1(t)
    t+1
end

function f2(t)
    t+sq2
end

function f3(t)
    t+sq3
end

# two dimensional manifolds

function f12(t,Δ)
    if 0 <= Δ <= 1/sq2
        return t + sqrt(1-Δ^2)
    end
    return Inf
end

function f23(t,Δ)
    if 0 <= Δ <= 1/sq3
        return t + sqrt(2-2Δ^2)
    end
    return Inf
end

function f13(t,Δ)
    if 0 <= Δ <= 2/sq3
        return t + sqrt(1-Δ^2/2)
    end
    return Inf
end

# Three dimensional manifolds
function f123(t,Δ1,Δ2)
    Δ² = Δ1^2+Δ2^2
    if Δ² <= 1 
        S = sqrt(1-Δ²)
        if 0 <= Δ2 <= Δ1 <= S
            return t + S
        end
    end
    return Inf   
end;

###################################################################
function paint(T::Array{Float64,2};bar=true,axis=true)
    n = size(T,1)-1
    Plots.contourf((0:n)/n,(0:n)/n,T',color=:turbo,aspect_ratio=:equal,clabels=!bar,cbar=bar,axis=axis,xlims=(0,1),legendfontsize=12,tickfontsize=12,levels=20,guidefontfamily="Times",tickfontfamily="Times")
    Plots.title!("Time field")
end;
